'use strict';

const assert = require('assert');
const is = require('is-type-of');
const tilNextMillis = Symbol('tilNextMillis');

/**
 * 雪花算法分布式唯一ID生成器
 * 每个机器号最高支持每秒64000（1000*2的6次方）个序列, 当秒序列不足时启用备份机器号, 若备份机器也不足时借用备份机器下一秒可用序列
 * JS的Number类型精度最高只有53bit，53 bits 趋势自增ID结构如下:
 * |00000000|00011111|11111111|11111111|11111111|11111111|11111111|11111111|
 * |-----------|##########41bit 毫秒秒级时间戳#################|-------------|
 * |------------------------------------------------6bit机器位|xxxxxx|------|
 * |-----------------------------------------------------6bit自增序列|xxxxxx|
 **/
class Snowflake {
  constructor(app, {
    // 初始偏移时间戳 (41bit，要求小于当前时间戳大于2000年时间戳)
    TWEPOCH = 1484201323141,
    // 机器id所占位数 (6bit, 支持最大机器数 2^6 = 64)
    WORKER_ID_BITS = 6,
    // 自增序列所占位数 (6bit, 支持最大每秒生成 2^6 = 64)
    SEQUENCE_BITS = 6,
    // 机器Id，0-63
    workerId = 1,
  }) {
    assert(TWEPOCH && WORKER_ID_BITS && SEQUENCE_BITS,
      `[egg-cute-snowflake] 'TWEPOCH: ${TWEPOCH}', 'WORKER_ID_BITS: ${WORKER_ID_BITS}', 'SEQUENCE_BITS: ${SEQUENCE_BITS}', 'workerId: ${workerId}', are required on config`);

    assert(is.long(TWEPOCH) && is.int32(WORKER_ID_BITS) && is.int32(SEQUENCE_BITS) && is.int32(workerId),
      `[egg-cute-snowflake] 'TWEPOCH: ${TWEPOCH}', 'WORKER_ID_BITS: ${WORKER_ID_BITS}', 'SEQUENCE_BITS: ${SEQUENCE_BITS}', 'workerId: ${workerId}', must be Integer on config`);

    assert((TWEPOCH < new Date().getTime() && TWEPOCH > 946656000000), '[egg-cute-snowflake] TWEPOCH value must be between 2000-01-01 and now.');

    assert(WORKER_ID_BITS + SEQUENCE_BITS === 12, '[egg-cute-snowflake] WORKER_ID_BITS + SEQUENCE_BITS must be equal 12');

    app.coreLogger.info('[egg-cute-snowflake] init workerId: %s', workerId);

    this.TWEPOCH = TWEPOCH;
    /** 机器标识最大值 63 */
    // eslint-disable-next-line no-bitwise
    this.MAX_WORKER_ID = -1 ^ (-1 << WORKER_ID_BITS);
    /** 机器id偏移位数 */
    this.WORKER_ID_SHIFT = SEQUENCE_BITS;
    /** 自增序列偏移位数 12 */
    this.TIMESTAMP_LEFT_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS;
    /** 自增序列最大值 63 */
    // eslint-disable-next-line no-bitwise
    this.SEQUENCE_MASK = -1 ^ (-1 << SEQUENCE_BITS);

    assert(workerId <= this.MAX_WORKER_ID && workerId >= 0, `[egg-cute-snowflake] worker Id can't be greater than ${this.MAX_WORKER_ID} or less than 0`);

    this.workerId = workerId;
    this.lastTimestamp = -1;
    this.sequence = 0;
  }

  nextId() {
    let timestamp = new Date().getTime();
    if (this.lastTimestamp === timestamp) {
      // eslint-disable-next-line no-bitwise
      this.sequence = (this.sequence + 1) & this.SEQUENCE_MASK;
      if (this.sequence === 0) {
        timestamp = this[tilNextMillis](this.lastTimestamp);
      }
    } else {
      this.sequence = 0;
    }

    assert(timestamp >= this.lastTimestamp, `[egg-cute-snowflake] lock moved backwards. Refusing to generate id for ${this.lastTimestamp - timestamp} milliseconds`);

    this.lastTimestamp = timestamp;
    // eslint-disable-next-line no-bitwise
    return ((timestamp - this.TWEPOCH) * (2 ** this.TIMESTAMP_LEFT_SHIFT)) + (this.workerId << this.WORKER_ID_SHIFT) + this.sequence;
  }

  [tilNextMillis](lastTimestamp) {
    let timestamp = new Date().getTime();
    while (timestamp <= lastTimestamp) {
      timestamp = new Date().getTime();
    }
    return timestamp;
  }
}

function create(config = {}, app) {
  return new Snowflake(app, config);
}

exports.app = app => {
  app.addSingleton('snowflake', create);
};
